// QFormConfiguration.h implementation of the QForm and 
//                      QFormConfiguration classes.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "QDSManager.h"
#include "QFormConfiguration.h"

//////////////////////////////////////////////////////////////////////
// Construction/destruction
//////////////////////////////////////////////////////////////////////

QForm::QForm( const CString &name_arg, QFormInput *target_arg, PtrList<QFormInput> &context_arg, QRuleSet *rule_set_arg )
{
  WIN32_GUFF();

  name = name_arg;
  target = target_arg;
  rule_set = rule_set_arg;

  POSITION pos = context_arg.GetHeadPosition();
  while ( pos != NULL )
    {
      context.AddTail( context_arg.GetNext( pos ) );
    }
}

QForm::~QForm()
{
  WIN32_GUFF();

	delete target;

	POSITION pos = context.GetHeadPosition();
  while ( pos != NULL )
		delete context.GetNext( pos );
}

//////////////////////////////////////////////////////////////////////
// Build a set of QFormFieldInputs that can be "filled in" to
// specify form search input
void QForm::BuildInputFields( QFormInput **target_ffi, PtrList<QFormInput> &context_ffi_list )
{
  WIN32_GUFF();

	//
	// Target field first 
	*target_ffi = target->CopySelf();

	//
	// Context fields
	POSITION pos = context.GetHeadPosition();
	while ( pos != NULL )
    context_ffi_list.AddTail( context.GetNext( pos )->CopySelf() );
}

//////////////////////////////////////////////////////////////////////
// Return the search classes that make up this form.
void QForm::Context( PtrList<QFormInput> &context_get )
{
  WIN32_GUFF();

  POSITION pos = context.GetHeadPosition();
  while ( pos != NULL )
    context_get.AddTail( context.GetNext( pos ) );
}

//////////////////////////////////////////////////////////////////////
// Construction/destruction
//////////////////////////////////////////////////////////////////////

QFormConfiguration::QFormConfiguration( QDSManager *ds_arg )
{
  ds = ds_arg;
}

QFormConfiguration::~QFormConfiguration()
{
  WIN32_GUFF();

  POSITION pos = form_list.GetHeadPosition();
  while ( pos != NULL )
    {
      delete form_list.GetNext( pos );
    }
  
  pos = rule_set_list.GetHeadPosition();
  while ( pos != NULL )
    {
      delete rule_set_list.GetNext( pos );
    }
}

//////////////////////////////////////////////////////////////////////
// Get a list of the names of all forms.
void QFormConfiguration::GetFormList( CStringList &form_name_list )
{
  WIN32_GUFF();

  POSITION pos = form_list.GetHeadPosition();
  while ( pos != NULL )
    {
      QForm *form = form_list.GetNext( pos );
      form_name_list.AddTail( form->Name() );
    }
}

//////////////////////////////////////////////////////////////////////
// Return the named form.
QForm *QFormConfiguration::GetForm( CString &form_name )
{
  WIN32_GUFF();

  POSITION pos = form_list.GetHeadPosition();
  while ( pos != NULL )
    {
      QForm *form = form_list.GetNext( pos );
      if ( form->Name() == form_name )
        {
          return form;
        }
    }

  return NULL;
}

//////////////////////////////////////////////////////////////////////
// Load form configuration.
// XXX: QFormConfiguration::LoadConfiguration() should return a QError
BOOL QFormConfiguration::LoadConfiguration( CString & config_dn, CString &host, int port )
{
  WIN32_GUFF();

	//
	// Bind to the default host
	char *tmp_host = strdup( host );
	LDAP *ld = ldap_init( ( LPCTSTR ) host, port );

	free( tmp_host );
  if ( ld == ( LDAP * ) NULL )
		return FALSE;

	// Set LDAP protocol version 3
  int version = LDAP_VERSION3;
  ldap_set_option( ld, LDAP_OPT_PROTOCOL_VERSION, &version );

  // Automatic referral chasing off
  ldap_set_option( ld, LDAP_OPT_REFERRALS, LDAP_OPT_OFF );

	// Try binding to directory server
	int ok = ldap_simple_bind_s( ld, NULL, NULL );
  if ( ok != LDAP_SUCCESS )
    return FALSE;

	//
	// Load rule sets
	if ( LoadRuleSets( ld, config_dn ) == FALSE )
		{
			return FALSE;
		}

	//
	// Search for forms beneath the configuration root entry
	if ( LoadForms( ld, config_dn ) == FALSE )
		{
			return FALSE;
		}

	ldap_unbind( ld );
	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// Load rule sets.
BOOL QFormConfiguration::LoadForms( LDAP *ld, CString base_dn )
{
	char *attrs[] =
		{ "cn", "pgInputList", "pgRuleSetName", "pgToolTip", "pgHelpString", NULL };

	LDAPMessage *lm, *lmi;
	char **vals;
	BerElement *ber;
	int ok;

	char *tmp_dn_str = strdup( base_dn );
	ok = ldap_search_ext_s
		( ld, tmp_dn_str, LDAP_SCOPE_ONELEVEL,
		  "(objectclass=pgSearchForm)",
			attrs, 0, NULL, NULL, 
			NULL, LDAP_NO_LIMIT, &lm );

	free( tmp_dn_str );

	if ( ok != LDAP_SUCCESS )
		{
#ifdef _DEBUG
      TRACE
				( "QFormConfiguration::LoadForms\
					\nldap_search_ext_s: %s\n",
					ldap_err2string( ok ) );
#endif

      return FALSE;
    }

	//
	// Attributes that we need to look for.
	QAttrT *cn_at = ds->Table()->str2at( CString( "cn" ) ),
		*il_at = ds->Table()->str2at( CString( "pgInputList" ) ),
		*rs_at = ds->Table()->str2at( CString( "pgRuleSetName" ) ),
		*tt_at = ds->Table()->str2at( CString( "pgToolTip" ) ),
		*hs_at = ds->Table()->str2at( CString( "pgHelpString" ) );

	for ( lmi = ldap_first_message( ld, lm ); lmi != NULL; lmi = ldap_next_message( ld, lmi ) )
		{
			if ( ldap_msgtype( lmi ) == LDAP_RES_SEARCH_ENTRY )
				{
					//
					// The form definition is loaded into these data variables...
					CString form_name, rs_name;
					CString tip_str, help_str;
					QFormInput *target_input = NULL;
					PtrList<QFormInput> context_input_list;

					//
		      // Get form entry DN
					char *form_dn_str;
					if ( ( form_dn_str = ldap_get_dn( ld, lmi ) ) == NULL )
						{
#ifdef _DEBUG
							TRACE
								( "QFormConfiguration::LoadForms\
									\nFailed to get DN from LDAP message whilst reading a form entry!\n" );
#endif
							continue;
						}

					//
					// Get the rule set first
					for ( char *a = ldap_first_attribute( ld, lmi, &ber );
								a != NULL;
								a = ldap_next_attribute( ld, lmi, ber ) )
						{
							QAttrT *this_at = ds->Table()->str2at( CString( a ) );
							if ( this_at == rs_at )
								{
									if ( ( vals = ldap_get_values( ld, lmi, a ) ) != NULL )
										{
											//
											// Each attribute in this entry should 
											// only have one value
											char *val = vals[0];

											if ( this_at == rs_at )
												{
													rs_name = val;
												}
										}

									ldap_value_free( vals );
								}

							ldap_memfree( a );
						}

					QRuleSet *rule_set = GetRuleSet( rs_name );
					if ( rule_set == NULL )
						{
#ifdef _DEBUG
							TRACE
								( "QFormConfiguration::LoadForms\
									\nInvalid rule set name '%s' in form entry:\
									\nDN: %s\n",
									rs_name, form_dn_str );

							ldap_memfree( form_dn_str );
							continue;
#endif
						}

					//
					// Parse remaining entry attributes
					for ( a = ldap_first_attribute( ld, lmi, &ber );
								a != NULL;
								a = ldap_next_attribute( ld, lmi, ber ) )
						{
							QAttrT *this_at = ds->Table()->str2at( CString( a ) );
							if ( ( vals = ldap_get_values( ld, lmi, a ) ) != NULL )
								{
									//
									// Each attribute in this entry should 
									// only have one value
									char *val = vals[0];

									if ( this_at == cn_at )
										{
											//
											// The name of the form
											form_name = val;
										}
									else if ( this_at == il_at )
										{
											//
											// The list of input fields
											QStringList input_name_list;
											CString il_val = val;

											input_name_list.SetSplitString( il_val, '$' );
											BOOL target = TRUE;

											POSITION il_pos = input_name_list.GetHeadPosition();
											while ( il_pos != NULL )
												{
													CString input_name = input_name_list.GetNext( il_pos );

													QFormInput *input =
														LoadFormInput( ld, form_dn_str, input_name, rule_set );

													if ( target == TRUE )
														{
															target_input = input;
															target = FALSE;
														}
													else 
														{
															context_input_list.AddTail( input );
														}
												}
										}
									else if ( this_at == tt_at )
										{
											//
											// Tip string
											tip_str = val;
										}
									else if ( this_at == hs_at )
										{
											//
											// Help string
											help_str = val;
										}

									ldap_value_free( vals );
								}

							ldap_memfree( a );
						}

					//
					// Check that we've got all the required info
					if ( form_name.IsEmpty() || target_input == NULL )
						{
#ifdef _DEBUG
							CString debug_msg;
							debug_msg.Format
								( "Failed to load form subtree at:\n%s\n", form_dn_str );
#endif
						}
					else
						//
						// We're OK so add the new form definition
						{
							AddForm( form_name, target_input, context_input_list, rs_name, tip_str, help_str );
						}

					ldap_memfree( form_dn_str );
				}
		}

	//
	// ALl done
	ldap_msgfree( lm );
	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// Load a form input.
QFormInput *QFormConfiguration::LoadFormInput
	( LDAP *ld, char *form_dn_str, CString &input_name, QRuleSet *rule_set )
{
	char *attrs[] = { "cn", "pgFieldList", "pgSearchClassName", NULL };

	QFormInput *fi = NULL;
	LDAPMessage *lm, *lmi;
	char **vals;
	BerElement *ber;
	int ok;

	CString filt_str = "(cn=" + input_name + ")";
	ok = ldap_search_ext_s
		( ld, (LPCTSTR) form_dn_str,
			LDAP_SCOPE_ONELEVEL,
		  const_cast<char *>( (LPCTSTR) filt_str ),
			attrs, 0, NULL, NULL, 
			NULL, LDAP_NO_LIMIT, &lm );

	if ( ok != LDAP_SUCCESS )
		{
#ifdef _DEBUG
			TRACE
				( "QFormConfiguration::LoadFormInput\
					\nldap_search_ext_s: %s\n",
					ldap_err2string( ok ) );
#endif

      return NULL;
    }

	//
	// Attributes that we need to look for.
	QAttrT *cn_at = ds->Table()->str2at( CString( "cn" ) ),
		*fl_at = ds->Table()->str2at( CString( "pgFieldList" ) ),
		*sc_at = ds->Table()->str2at( CString( "pgSearchClassName" ) );

	for ( lmi = ldap_first_message( ld, lm ); lmi != NULL; lmi = ldap_next_message( ld, lmi ) )
		{
      if ( ldap_msgtype( lmi ) == LDAP_RES_SEARCH_ENTRY )
				{
					//
					// The form input definition is loaded into these data variables...
					CString input_name, sc_name;
					PtrList<QFormField> field_list;

					//
		      // Get form entry DN
					char *input_dn_str;
					if ( ( input_dn_str = ldap_get_dn( ld, lmi ) ) == NULL )
						{
#ifdef _DEBUG
							TRACE
								( "QFormConfiguration::LoadFormInput\
									\nFailed to get DN from LDAP message whilst reading a form input entry!\n" );
#endif
							ldap_msgfree( lm );
							return NULL;
						}

					//
					// Parse form entry attributes
					for ( char *a = ldap_first_attribute( ld, lmi, &ber );
								a != NULL;
								a = ldap_next_attribute( ld, lmi, ber ) )
						{
							QAttrT *this_at = ds->Table()->str2at( CString( a ) );
							if ( ( vals = ldap_get_values( ld, lmi, a ) ) != NULL )
								{
									//
									// Each attribute in this entry should 
									// only have one value
									char *val = vals[0];

									if ( this_at == cn_at )
										// Input name
										{
											input_name = val;
										}
									else if ( this_at == fl_at )
										// Field list
										{
											//
											// The list of input fields
											QStringList field_name_list;
											CString fl_val = val;

											field_name_list.SetSplitString( fl_val, '$' );

											POSITION fl_pos = field_name_list.GetHeadPosition();
											while ( fl_pos != NULL )
												{
													CString field_name = field_name_list.GetNext( fl_pos );
													QFormField *field =	LoadFormField( ld, input_dn_str, field_name );

													if ( field != NULL )
														field_list.AddTail( field );
												}
										}
									else if ( this_at == sc_at )
										//
										// Search class name
										{
											sc_name = val;
										}

									ldap_value_free( vals );
								}

							ldap_memfree( a );
						}

					//
					// Check that we've got all the required info
					QSearchClass *sc = rule_set->GetNamedClass( sc_name );
					if ( input_name.IsEmpty() || sc == NULL || field_list.GetCount() == 0 )
						{
#ifdef _DEBUG
							TRACE
								( "QFormConfiguration::LoadFormInput\
									\nFailed to load form input subtree at:\
									\n%s\n",
									input_dn_str );
#endif
						}
					else
						//
						// We're OK so add the new form definition
						{
							fi = new QFormInput( input_name, sc, field_list );
						}

					ldap_memfree( input_dn_str );
				}
		}

	//
	// All done
	ldap_msgfree( lm );
	return fi;
}

//////////////////////////////////////////////////////////////////////
// Load form field
QFormField *QFormConfiguration::LoadFormField( LDAP *ld, char *input_dn_str, CString &field_name )
{
	char *attrs[] = { "cn", "pgFiltAttrList", NULL };

	QFormField *ff = NULL;
	LDAPMessage *lm, *lmi;
	char **vals;
	BerElement *ber;
	int ok;

	CString filt_str = "(cn=" + field_name + ")";
	ok = ldap_search_ext_s
		( ld, (LPCTSTR) input_dn_str,
			LDAP_SCOPE_ONELEVEL,
		  const_cast<char *>( (LPCTSTR) filt_str ),
			attrs, 0, NULL, NULL, 
			NULL, LDAP_NO_LIMIT, &lm );

	if ( ok != LDAP_SUCCESS )
		{
#ifdef _DEBUG
			TRACE
				( "QFormConfiguration::LoadFormField\
					ldap_search_ext_s: %s\n",
					ldap_err2string( ok ) );
#endif

      return NULL;
    }

	//
	// Attributes that we need to look for.
	QAttrT *cn_at = ds->Table()->str2at( CString( "cn" ) ),
		*fl_at = ds->Table()->str2at( CString( "pgFiltAttrList" ) );

	for ( lmi = ldap_first_message( ld, lm ); lmi != NULL; lmi = ldap_next_message( ld, lmi ) )
		{
      if ( ldap_msgtype( lmi ) == LDAP_RES_SEARCH_ENTRY )
				{
					//
					// The form input definition is loaded into these data variables...
					CString field_name, default_value = "";
					PtrList<QAttrT> filt_at_list;

					//
		      // Get form entry DN
					char *field_dn_str;
					if ( ( field_dn_str = ldap_get_dn( ld, lmi ) ) == NULL )
						{
#ifdef _DEBUG
							TRACE
								( "QFormConfiguration::LoadFormField\
									\nFailed to get DN from LDAP message whilst reading a form field entry!\n" );
#endif

							ldap_msgfree( lm );
							return NULL;
						}

					//
					// Parse form entry attributes
					for ( char *a = ldap_first_attribute( ld, lmi, &ber );
								a != NULL;
								a = ldap_next_attribute( ld, lmi, ber ) )
						{
							QAttrT *this_at = ds->Table()->str2at( CString( a ) );
							if ( ( vals = ldap_get_values( ld, lmi, a ) ) != NULL )
								{
									//
									// Each attribute in this entry should 
									// only have one value
									char *val = vals[0];

									if ( this_at == cn_at )
										// Input name
										{
											field_name = val;
										}
									else if ( this_at == fl_at )
										{
											// The list of input fields
											QStringList at_name_list;
											CString fl_val = val;

											at_name_list.SetSplitString( fl_val, '$' );

											POSITION at_pos = at_name_list.GetHeadPosition();
											while ( at_pos != NULL )
												{
													CString at_name = at_name_list.GetNext( at_pos );
													QAttrT *at = ds->Table()->str2at( at_name );

													if ( at != NULL )
														filt_at_list.AddTail( at );
												}
										}

									ldap_value_free( vals );
								}

							ldap_memfree( a );
						}

					//
					// Check that we've got all the required info
					if ( field_name.IsEmpty() || filt_at_list.GetCount() == 0 )
						{
#ifdef _DEBUG
							CString debug_msg;
							debug_msg.Format
								( "Failed to load form field at:\n%s\n", field_dn_str );
#endif
						}
					else
						//
						// We're OK so add the new form definition
						{
							ff = new QFormField( field_name, default_value, filt_at_list );
						}

					ldap_memfree( field_dn_str );
				}
		}

	//
	// All done
	ldap_msgfree( lm );
	return ff;
}

//////////////////////////////////////////////////////////////////////
// Load rule sets.
BOOL QFormConfiguration::LoadRuleSets( LDAP *ld, CString base_dn )
{
	char *attrs[] = { "cn", NULL} ;

	LDAPMessage *lm, *lmi;
	char **vals;
	BerElement *ber;
	int ok;

	ok = ldap_search_ext_s
		( ld, const_cast<char *>( (LPCTSTR) base_dn ), LDAP_SCOPE_ONELEVEL,
		  "(objectclass=pgRuleSet)",
			attrs, 0, NULL, NULL, 
			NULL, LDAP_NO_LIMIT, &lm );

	if ( ok != LDAP_SUCCESS )
		{
#ifdef _DEBUG
			TRACE
				( "QFormConfiguration::LoadRuleSets\
					\nldap_search_ext_s: %s\n",
					ldap_err2string( ok ) );
#endif

      return FALSE;
    }

	//
	// Attributes that we need to look for.
	QAttrT *cn_at = ds->Table()->str2at( CString( "cn" ) );
	for ( lmi = ldap_first_message( ld, lm ); lmi != NULL; lmi = ldap_next_message( ld, lmi ) )
		{
      if ( ldap_msgtype( lmi ) == LDAP_RES_SEARCH_ENTRY )
				{
					//
		      // Get and load the search class
					char *rs_dn_str;
					if ( ( rs_dn_str = ldap_get_dn( ld, lmi ) ) == NULL )
						{
#ifdef _DEBUG
							TRACE
								( "QFormConfiguration::LoadRuleSets\
									\nFailed to get DN from LDAP message whilst reading rule set entries!\n" );
#endif
							continue;
						}

					// Load the search classes belonging to this form
					PtrList<QSearchClass> sc_list;
					LoadSearchClasses( ld, CString( rs_dn_str ), sc_list );
					ldap_memfree( rs_dn_str );

					// Get rule set name
					CString rs_name; 

					// Parse search class entry attributes
					for ( char *a = ldap_first_attribute( ld, lmi, &ber );
								a != NULL;
								a = ldap_next_attribute( ld, lmi, ber ) )
						{
							QAttrT *this_at = ds->Table()->str2at( CString( a ) );
							if ( ( vals = ldap_get_values( ld, lmi, a ) ) != NULL )
								{
									for ( int i = 0; vals[ i ] != NULL; i++ )
										{
											if ( this_at == cn_at )
												// Search class name
												{
													rs_name = vals[i];
												}
										}

									ldap_value_free( vals );
								}

							ldap_memfree( a );
						}

					//
					// Add the rule set
					AddRuleSet( rs_name, sc_list );
					sc_list.RemoveAll();
				}
		}

	//
	// Free the LDAP result
	ldap_msgfree( lm );
	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// Load search classes from the directory.
BOOL QFormConfiguration::LoadSearchClasses( LDAP *ld, CString dn_str, PtrList<QSearchClass> &sc_list )
{
	char *attrs[] =
		{ "cn", "pgObjectClassDefinition", "pgSearchClassList", "pgLDAPURI", NULL } ;

	LDAPMessage *lm, *lmi;
	char **vals;
	char *a;
	BerElement *ber;
	int ok;

	ok = ldap_search_ext_s
		( ld, const_cast<char *>( (LPCTSTR) dn_str ), LDAP_SCOPE_ONELEVEL,
		  "(objectclass=pgSearchClass)",
			attrs, 0, NULL, NULL, 
			NULL, LDAP_NO_LIMIT, &lm );

	if ( ok != LDAP_SUCCESS )
		return FALSE;

	//
	// Attributes that we need to look for.
	QAttrT *cn_at = ds->Table()->str2at( CString( "cn" ) ),
		*oc_def_at = ds->Table()->str2at( CString( "pgObjectClassDefinition" ) ),
		*sc_list_at = ds->Table()->str2at( CString( "pgSearchClassList" ) ),
		*ldap_uri_at = ds->Table()->str2at( CString( "pgLDAPURI" ) );

	for ( lmi = ldap_first_message( ld, lm ); lmi != NULL; lmi = ldap_next_message( ld, lmi ) )
		{
      if ( ldap_msgtype( lmi ) == LDAP_RES_SEARCH_ENTRY )
				{
					//
		      // Get and load the search class
					char *sc_dn_str;
					if ( ( sc_dn_str = ldap_get_dn( ld, lmi ) ) == NULL )
						{
#ifdef _DEBUG
							TRACE
								( "QFormConfiguration::LoadSearchClasses\
									\nFailed to get DN from LDAP message whilst reading search classes!\n" );
#endif
							continue;
						}

					// Iterate through each attribute in the entry
					CString sc_name; 
					PtrList<QOC> oc_list;
					CString ldap_uri;

					// Reset search class definition data structures 
					sc_name.Empty();
					if ( oc_list.GetCount() > 0 )
						oc_list.RemoveAll();

					// Parse search class entry attributes
					for ( a = ldap_first_attribute( ld, lmi, &ber );
								a != NULL;
								a = ldap_next_attribute( ld, lmi, ber ) )
						{
							QAttrT *this_at = ds->Table()->str2at( CString( a ) );
							if ( ( vals = ldap_get_values( ld, lmi, a ) ) != NULL )
								{
									if ( this_at == cn_at )
										// Search class name
										{
											sc_name = vals[0];
										}
									else if ( this_at == ldap_uri_at )
										{
											ldap_uri = vals[0];
										}
									else if ( this_at == oc_def_at )
										// Object class definition
										{
											for ( int i = 0; vals[ i ] != NULL; i++ )
												{
													QOC *oc = ds->Table()->str2oc( CString( vals[i] ) );
													if ( oc != NULL )
														oc_list.AddTail( oc );
#ifdef _DEBUG
													else
														{
															TRACE
																( "QFormConfiguration::LoadSearchClasses\
																	Unrecognized object class in search class entry:\
																	\nOC: %s\
																	\nDN: %s\n",
																	vals[i], sc_dn_str );
														}
#endif
												}
										}

									ldap_value_free( vals );
								}

							ldap_memfree( a );
						}

					if ( ber != NULL )
						ber_free( ber, 0 );

					// Add the search class
					QSearchClass *sc = NULL;
					if ( oc_list.GetCount() > 0 )
						{
							sc = new QSearchClass( ds->Table(), sc_name, oc_list );
						}
					else if ( ldap_uri.IsEmpty() == FALSE )
						{
							sc = new QSearchClass( ds, sc_name, ldap_uri );
						}
					
					if ( sc == NULL )
						{
#ifdef _DEBUG
							CString debug_msg;
							debug_msg.Format
								( "Badly defined search class in entry:\n%s\n", sc_dn_str );
#endif
							// Badly defined search class!
							ldap_memfree( sc_dn_str );
							continue;
						}

					sc_list.AddTail( sc );
					ldap_memfree( sc_dn_str );
				}
		}

	//
	// Need to loop through results again to get search class hierarchy. Need
	// to make a second pass because all search classes need to be defined beforehand
	// in order to load the information.
	POSITION sc_pos = sc_list.GetHeadPosition();
	for ( lmi = ldap_first_message( ld, lm ); lmi != NULL; lmi = ldap_next_message( ld, lmi ) )
		{
			QSearchClass *sc_parent;
			if ( sc_pos != NULL )
				 sc_parent = sc_list.GetNext( sc_pos );

      if ( ldap_msgtype( lmi ) == LDAP_RES_SEARCH_ENTRY )
				{
					//
		      // Get and load the search class
					char *sc_dn_str;
					if ( ( sc_dn_str = ldap_get_dn( ld, lmi ) ) != NULL )
						ldap_memfree( sc_dn_str );
					else
						{
#ifdef _DEBUG
							TRACE
								( "QFormConfiguration::LoadSearchClasses\
									\nFailed to get DN from LDAP message whilst reading search classes!\n" );
#endif
							continue;
						}

					//
					// Get search class hierarchy information (attr pgSearchClassList)
					for ( a = ldap_first_attribute( ld, lmi, &ber );
								a != NULL;
								a = ldap_next_attribute( ld, lmi, ber ) )
						{
							QAttrT *this_at = ds->Table()->str2at( CString( a ) );
							if ( this_at == sc_list_at )
								{
									if ( ( vals = ldap_get_values( ld, lmi, a ) ) != NULL )
										{
											for ( int i = 0; vals[ i ] != NULL; i++ )
												{
													QStringList sc_name_list;
													BOOL got_it = FALSE;
													QSearchClass *sc_child;
													
													sc_name_list.SetSplitString( CString( vals[i] ), '$' );

													POSITION name_pos = sc_name_list.GetHeadPosition();
													while ( name_pos != NULL )
														{
															//
															// Get the named search class
															CString sc_name = sc_name_list.GetNext( name_pos );
															POSITION child_sc_pos = sc_list.GetHeadPosition();

															while ( child_sc_pos != NULL )
																{
																	sc_child = sc_list.GetNext( child_sc_pos );
																	if ( sc_child->Name() == sc_name )
																		{
																			got_it = TRUE;
																			break;
																		}
																}

															if ( got_it == TRUE )
																break;
														}

													//
													// Add the identified search class
													if ( got_it == TRUE )
														sc_parent->AddSearchClass( sc_child );
												}

											ldap_value_free( vals );
										}
								}

							ldap_memfree( a );
						}

					if ( ber != NULL )
						ber_free( ber, 0 );
				}
		}

	//
	// All done...
	ldap_msgfree( lm );
	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// Add a form to the configuration.
BOOL QFormConfiguration::AddForm
  ( CString &name,
		QFormInput *target,
		PtrList<QFormInput> &context,
		CString &rule_set_name,
		CString &tip_str,
		CString &help_str )
{
  WIN32_GUFF();

  //
  // Get the rule set
  QRuleSet *rs = GetRuleSet( rule_set_name );
  if ( rs == NULL )
    {
#ifdef _DEBUG
      TRACE
        ( "QFormConfiguration::AddForm\
					\nFailed to find named rule set when defining form\
					\nRule Set: %s\
					\nForm: %s\n",
					rule_set_name,
					name );
#endif
      return FALSE;
    }

  //
  // Create the new form specification and add it to the list
  QForm *new_form = new QForm( name, target, context, rs );
  form_list.AddTail( new_form );

	//
	// Set the help and tip strings
	new_form->TipString() = tip_str;
	new_form->HelpString() = help_str;

  return TRUE;
}

//////////////////////////////////////////////////////////////////////
// Add a form to the configuration.
BOOL QFormConfiguration::AddRuleSet( CString &name, PtrList<QSearchClass> &class_list )
{ 
  WIN32_GUFF();

  //
  // Check that a rule set with the given name doesn't already exist
  QRuleSet *rs = GetRuleSet( name );
  if ( rs != NULL )
    {
#ifdef _DEBUG
      TRACE
				( "QFormConfiguration::AddRuleSet\
					\nMore than one definition of rule set found.\
					\nRule Set: %s",
					name );
#endif
      return FALSE;
    }
  
  //
  // Create the new rule set and add to list
  rs = new QRuleSet( name, class_list );
  rule_set_list.AddTail( rs );
  return TRUE;
}

//////////////////////////////////////////////////////////////////////
// Get the named rule set.
QRuleSet* QFormConfiguration::GetRuleSet( CString &name )
{
  WIN32_GUFF();

  POSITION pos = rule_set_list.GetHeadPosition();
  while ( pos != NULL )
    {
      QRuleSet *rs = rule_set_list.GetNext( pos );
      if ( rs->Name() == name )
        return rs;
    }

  return NULL;
}

